function x = gpoissonrnd(alpha, beta, dims, epsilon)
% # Generalized Poisson random variable generator
% ## Usage
% * `x = gpoissonrnd()`
% * `x = gpoissonrnd(alpha, beta, dims, epsilon)`
%
% ## Description
% Generates random samples distributed according to the generalized Poisson
% distribution [1] defined as
%
%     p(x) = alpha*(alpha+beta*x)^(x-1)*exp(-alpha-beta*x)/x!
%
% Notes:
% * If `beta = 0`, the generalized Poisson distribution coincides with the
%   regular Poisson distribution.
% * Since the mixture distribution to generate generalized Poisson random
%   variables is hard to derive, this function instead calculates the
%   weights and uses categorical sampling to sample. This requires
%   truncation at some upper bound `xmax`, which is calculated based on the
%   mean and covariance and the remaining tail probability is assigned to 
%   the maximum value `xmax+1`. If the tail probability exceed the 
%   threshold `epsilon`, a warning is shown.
%
% ## Input
% * `alpha`: Location parameter (default: `1`).
% * `beta`: Skewness/scale parameter (default: `0`).
% * `dims`: Dimensions (default: `[1, 1]`).
% * `epsilon`: Tail probability threshold (default: `1e-2`).
%
% ## Outut
% * `x`: Matrix of random samples of dimension `dims`.
%
% ## References
% 1. P. C. Consul and G. C. Jain, “A generalization of the Poisson 
%    distribution,” Technometrics, vol. 15, no. 4, pp. 791–799, 1973.
% 
% ## Authors
% 2019-present -- Roland Hostettler

%{
% This file is free software: you can redistribute it and/or modify it 
% under the terms of the GNU General Public License as published by the 
% Free Software Foundation, either version 3 of the License, or (at your
% option) any later version.
% 
% This file is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
% FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
% more details.
% 
% You should have received a copy of the GNU General Public License along 
% with this file. If not, see <http://www.gnu.org/licenses/>.
%}

% TODO:
% * Calculation of px breaks down if alpha, beta are not scalars and/or
%   same dimension as 0:xmax; should be addressed.

    %% Defaults
    narginchk(0, 4);
    if nargin < 1 || isempty(alpha)
        alpha = 1;
    end
    if nargin < 2 || isempty(beta)
        beta = 0;
    end
    if nargin < 3 || isempty(dims)
        % TODO: Adjust this when we've replaced catrand with our own catrnd
        dims = [1, 1];
    end
    if nargin < 4 || isempty(epsilon)
        epsilon = 1e-2;
    end
    
    %% Generate samples
    % Find the upper limit for calculating the the probabilities. Use the
    % mean and variance to estimate how far the significant portion of the
    % right tail extends.
    m = alpha./(1-beta);
    mmax = max(m(:));
    xmax = mmax+10*sqrt(mmax/(1-beta).^2);
    
    % Calculate probabilities
    px = exp(loggpoissonpdf(0:xmax, alpha, beta));
    ptail = 1-sum(px, 2);
    if ptail >= epsilon
        warning('Unaccounted tail probability exceeds %.2e, consider increasing xmax.', epsilon);
    end
    px = [px, ptail];
    
    x = zeros(dims);
    for ix = 1:dims(1)
        % Sample from categorical distribution defined by probabilities
        x(ix, :) = catrnd(px(ix, :), [1, dims(2)])-1;
    end
end
