function lpx = loggpoissonpdf(x, alpha, beta)
% # Generalized Poisson log-pdf
% ## Usage
% * `lpx = loggpoissonpdf(x)`
% * `lpx = loggpoissonpdf(x, alpha, beta)`
%
% ## Description
% Logarithm of the generalized Poisson probability density function
% according to [1] and given by
%
%     p(x) = alpha*(alpha+beta*x)^(x-1)*exp(-alpha-beta*x)/x!
% 
% Note: If `beta = 0`, the generalized Poisson distribution coincides with
% the regular Poisson distribution.
%
% ## Input
% * `x`: Draws of the random variable.
% * `alpha`: Location parameter (default: `1`).
% * `beta`: Skewness/scale parameter (default: `0`).
%
% ## Output
% * `lpx`: Logarithm of p(x).
%
% ## References
% 1. P. C. Consul and G. C. Jain, “A generalization of the Poisson 
%    distribution,” Technometrics, vol. 15, no. 4, pp. 791–799, 1973.
% 
% ## Author
% 2019-present -- Roland Hostettler

%{
% This file is free software: you can redistribute it and/or modify it 
% under the terms of the GNU General Public License as published by the 
% Free Software Foundation, either version 3 of the License, or (at your
% option) any later version.
% 
% This file is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
% FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
% more details.
% 
% You should have received a copy of the GNU General Public License along 
% with this file. If not, see <http://www.gnu.org/licenses/>.
%}

    %% Defaults
    narginchk(1, 3);
    if nargin < 2 || isempty(alpha)
        alpha = 1;
    end
    if nargin < 3 || isempty(beta)
        beta = 0;
    end
    
    %% Log-pdf
    lpx = log(alpha) + (x-1).*log(alpha + beta.*x) - (alpha + beta.*x) - logfactorial(x);
    lpx(isnan(lpx)) = -Inf;
end
